/*=============================================================================
	UnPrim.cpp: Unreal primitive functions.

	Copyright 1997 Epic MegaGames, Inc. This software is a trade secret.
	Compiled with Visual C++ 4.0. Best viewed with Tabs=4.

	Revision history:
		* Created by Tim Sweeney
=============================================================================*/

#include "Unreal.h"

/*----------------------------------------------------------------------------
	UPrimitive object implementation.
----------------------------------------------------------------------------*/

IMPLEMENT_CLASS(UPrimitive);

/*----------------------------------------------------------------------------
	UPrimitive collision checking.
----------------------------------------------------------------------------*/

//
// GetBoundingBox.
// Treats the primitive as a cyllinder.
//
FBoundingBox UPrimitive::GetBoundingBox( const AActor *Owner ) const
{
	guard(UPrimitive::GetBoundingBox);
	if( Owner )
	{
		FVector Extent( Owner->CollisionRadius+4, Owner->CollisionRadius+4, Owner->CollisionHeight+4 );
		return FBoundingBox( Owner->Location - Extent, Owner->Location + Extent );
	}
	else
	{
		return LocalBound;
	}
	unguard;
}

//
// PointCheck.
// Treats the primitive as a cyllinder.
//
INT UPrimitive::PointCheck
(
	FCheckResult	&Result,
	AActor			*Owner,
	const FVector	&Point,
	DWORD			ExtraNodeFlags
)
{
	guard(UPrimitive::PointCheck);
	if( Owner ) return
	(	Square(Owner->Location.Z-Point.Z)                                   > Square(Owner->CollisionHeight)
	||  Square(Owner->Location.X-Point.X)+Square(Owner->Location.Y-Point.Y) > Square(Owner->CollisionRadius) );
	else return 1;
	unguard;
}

//
// LineCheck.
// Treats the primitive as a cyllinder.
//
INT UPrimitive::LineCheck
(
	FCheckResult	&Result,
	AActor			*Owner,
	const FVector	&Start,
	const FVector	&End,
	DWORD ExtraNodeFlags
)
{
	guard(UPrimitive::LineCheck);
	return BoxLineCheck( Result, Owner, Start, End, 0, 0, ExtraNodeFlags );
	unguard;
}

//
// BoxPointCheck.
// Treats the primitive as a cyllinder.
//
INT UPrimitive::BoxPointCheck
(
	FCheckResult	&Result,
	AActor			*Owner,
	const FVector   &Point,
	FLOAT           Radius,
	FLOAT           Height,
	DWORD           ExtraNodeFlags
)
{
	guard(UPrimitive::BoxPointCheck);
	if( Owner ) return
	(	Square(Owner->Location.Z-Point.Z)                                   > Square(Owner->CollisionHeight+Height)
	||	Square(Owner->Location.X-Point.X)+Square(Owner->Location.Y-Point.Y) > Square(Owner->CollisionRadius+Radius) );
	else return 1;
	unguard;
}

//
// BoxLineCheck.
// Treats the primitive as a cyllinder.
//
INT UPrimitive::BoxLineCheck
(
	FCheckResult	&Result,
	AActor			*Owner,
	const FVector   &Start,
	const FVector   &End,
	FLOAT           Radius,
	FLOAT           Height,
	DWORD           ExtraNodeFlags
)
{
	guard(UPrimitive::BoxLineCheck);

	if( !Owner )
		return 1;

	// Treat this actor as a cyllinder.
	FLOAT NetRadius = Radius + Owner->CollisionRadius;
	FLOAT NetHeight = Height + Owner->CollisionHeight;

	// Quick X reject.
	FLOAT MaxX = Owner->Location.X + NetRadius;
	if( Start.X>MaxX && End.X>MaxX )
		return 1;
	FLOAT MinX = Owner->Location.X - NetRadius;
	if( Start.X<MinX && End.X<MinX )
		return 1;

	// Quick Y reject.
	FLOAT MaxY = Owner->Location.Y + NetRadius;
	if( Start.Y>MaxY && End.Y>MaxY )
		return 1;
	FLOAT MinY = Owner->Location.Y - NetRadius;
	if( Start.Y<MinY && End.Y<MinY )
		return 1;

	// Quick Z reject.
	FLOAT TopZ = Owner->Location.Z + NetHeight;
	if( Start.Z>TopZ && End.Z>TopZ )
		return 1;
	FLOAT BotZ = Owner->Location.Z - NetHeight;
	if( Start.Z<BotZ && End.Z<BotZ )
		return 1;

	// Clip to top of cyllinder.
	FLOAT T0=0.0, T1=1.0;
	if( Start.Z>TopZ && End.Z<TopZ )
	{
		FLOAT T = (TopZ - Start.Z)/(End.Z - Start.Z);
		if( T > T0 )
		{
			T0 = T;
			Result.Normal = FVector(0,0,1);
		}
	}
	else if( Start.Z<TopZ && End.Z>TopZ )
		T1 = ::Min( T1, (TopZ - Start.Z)/(End.Z - Start.Z) );

	// Clip to bottom of cyllinder.
	if( Start.Z<BotZ && End.Z>BotZ )
	{
		FLOAT T = (BotZ - Start.Z)/(End.Z - Start.Z);
		if( T > T0 )
		{
			T0 = T;
			Result.Normal = FVector(0,0,-1);
		}
	}
	else if( Start.Z>BotZ && End.Z<BotZ )
		T1 = ::Min( T0, (BotZ - Start.Z)/(End.Z - Start.Z) );
	
	// Reject.
	if( T0 >= T1 )
		return 1;

	// Test setup.
	FLOAT   Kx        = Start.X - Owner->Location.X;
	FLOAT   Ky        = Start.Y - Owner->Location.Y;

	// 2D circle clip about origin.
	FLOAT   Vx        = End.X   - Start.X;
	FLOAT   Vy        = End.Y   - Start.Y;
	FLOAT   A         = Vx*Vx + Vy*Vy;
	FLOAT   B         = 2.0 * (Kx*Vx + Ky*Vy);
	FLOAT   C         = Kx*Kx + Ky*Ky - NetRadius*NetRadius;
	FLOAT   Discrim   = B*B - 4.0*A*C;

	// No intersection if discriminant is negative.
	if( Discrim < 0 )
		return 1;

	// Unstable intersection if velocity is tiny.
	if( A < Square(0.0001) )
	{
		// Outside.
		if( C > 0 )
			return 1;
	}
	else
	{
		// Compute intersection times.
		Discrim   = sqrt(Discrim);
		FLOAT R2A = 0.5/A;
		T1        = ::Min( T1, +(Discrim-B) * R2A );
		FLOAT T   = -(Discrim+B) * R2A;
		if( T > T0 )
		{
			T0 = T;
			Result.Normal   = (Start + (End-Start)*T0 - Owner->Location);
			Result.Normal.Z = 0;
			Result.Normal.Normalize();
		}
		if( T0 >= T1 )
			return 1;
	}
	Result.Time      = ::Max(T0-0.001,0.0);
	Result.Location  = Start + (End-Start) * Result.Time;
	Result.Actor     = Owner;
	Result.Primitive = NULL;

	return 0;
	unguard;
}

/*----------------------------------------------------------------------------
	The End.
----------------------------------------------------------------------------*/
